@@ -10,6 +10,7 @@ import re
 import struct
 import sys
 from collections import defaultdict
+from typing import NamedTuple
 
 import numpy
 import xxhash
@@ -650,7 +651,7 @@ in file {self.file.file_path}"""
         ]
         self.deserialize_page_decompressed_buffer(destination, field_metadata)
 
-    def _expected_array_length_and_starts(
+    def _expected_array_length_starts_dtype(
         self, col_idx, cluster_start, cluster_stop, missing_element_padding=0
     ):
         """
@@ -660,7 +661,7 @@ in file {self.file.file_path}"""
             cluster_stop (int): The first cluster to exclude (i.e. one greater than the last cluster to include).
             missing_element_padding (int): Number of padding elements to add at the start of the array.
 
-        Returns the expected length of the array over the given cluster range, including padding, and also the start indices of each cluster.
+        Returns the expected length of the array over the given cluster range (including padding), the start indices of each cluster, and the dtype of the array.
         """
         field_metadata = self.get_field_metadata(col_idx)
         if field_metadata.dtype_byte in uproot.const.rntuple_index_types:
@@ -690,7 +691,7 @@ in file {self.file.file_path}"""
             starts.append(total_length)
             total_length += cluster_length
 
-        return total_length, starts
+        return total_length, starts, field_metadata.dtype_result
 
     def read_cluster_range(
         self,
@@ -699,6 +700,7 @@ in file {self.file.file_path}"""
         cluster_stop,
         missing_element_padding=0,
         array_cache=None,
+        access_log=None,
     ):
         """
         Args:
@@ -707,14 +709,31 @@ in file {self.file.file_path}"""
             cluster_stop (int): The first cluster to exclude (i.e. one greater than the last cluster to include).
             missing_element_padding (int): Number of padding elements to add at the start of the array.
             array_cache (None, or MutableMapping): Cache of arrays. If None, do not use a cache.
+            access_log (None or object with a ``__iadd__`` method): If an access_log is
+                provided, e.g. a list, cluster reads are tracked inside this reference.
 
         Returns a numpy array with the data from the column.
         """
+        if access_log is not None:
+            if not hasattr(access_log, "__iadd__"):
+                raise ValueError(f"{access_log=} needs to implement '__iadd__'.")
+            access_log += [
+                Accessed(
+                    column_index=col_idx,
+                    cluster_start=int(cluster_start),
+                    cluster_stop=int(cluster_stop),
+                    field_id=self.column_records[col_idx].field_id,
+                    field_name=self.field_records[
+                        self.column_records[col_idx].field_id
+                    ].field_name,
+                )
+            ]
+
         field_metadata = self.get_field_metadata(col_idx)
-        total_length, starts = self._expected_array_length_and_starts(
+        total_length, starts, dtype = self._expected_array_length_starts_dtype(
             col_idx, cluster_start, cluster_stop, missing_element_padding
         )
-        res = numpy.empty(total_length, field_metadata.dtype_result)
+        res = numpy.empty(total_length, dtype)
         # Initialize the padding elements. Note that it might be different from missing_element_padding
         # because for offsets there is an extra zero added at the start.
         assert len(starts) > 0, "The cluster range is invalid"
@@ -726,7 +745,7 @@ in file {self.file.file_path}"""
                 cluster_idx,
                 col_idx,
                 field_metadata,
-                destination=res[starts[i] : stop],
+                destination=res[starts[i] : stop].view(field_metadata.dtype),
                 array_cache=array_cache,
             )
 
@@ -964,7 +983,7 @@ in file {self.file.file_path}"""
             n_padding = self.column_records[key_nr].first_element_index
             n_padding -= cluster_starts[start_cluster_idx]
             n_padding = max(n_padding, 0)
-            total_length, starts = self._expected_array_length_and_starts(
+            total_length, starts, _ = self._expected_array_length_starts_dtype(
                 ncol, start_cluster_idx, stop_cluster_idx, n_padding
             )
             field_metadata = self.get_field_metadata(ncol)
@@ -1169,6 +1188,8 @@ in file {self.file.file_path}"""
             "std::string"
         ):
             dtype_result = dtype
+        elif dtype_byte in uproot.const.rntuple_custom_float_types:
+            dtype_result = numpy.float32
         else:
             dtype_result = numpy.result_type(*alt_dtype_list)
         field_metadata = FieldClusterMetadata(
@@ -1737,6 +1758,8 @@ class RField(uproot.behaviors.RNTuple.HasFields):
         interpreter="cpu",
         backend="cpu",
         ak_add_doc=False,
+        virtual=False,
+        access_log=None,
         # For compatibility reasons we also accepts kwargs meant for TTrees
         interpretation=None,
         interpretation_executor=None,
@@ -1763,6 +1786,11 @@ class RField(uproot.behaviors.RNTuple.HasFields):
                 to the Awkward ``__doc__`` parameter of the array.
                 if dict = {key:value} and ``library="ak"``, add the RField ``value`` to the
                 Awkward ``key`` parameter of the array.
+            virtual (bool): If True, return virtual Awkward arrays, meaning that the data will not be
+                loaded into memory until it is accessed.
+            access_log (None or object with a ``__iadd__`` method): If an access_log is
+                provided, e.g. a list, all materializations of the arrays are
+                tracked inside this reference. Only applies if ``virtual=True``.
             interpretation (None): This argument is not used and is only included for now
                 for compatibility with software that was used for :doc:`uproot.behaviors.TBranch.TBranch`. This argument should not be used
                 and will be removed in a future version.
@@ -1792,6 +1820,8 @@ class RField(uproot.behaviors.RNTuple.HasFields):
             interpreter=interpreter,
             backend=backend,
             ak_add_doc=ak_add_doc,
+            virtual=virtual,
+            access_log=access_log,
         )
         if self.name in arrays.fields:
             arrays = arrays[self.name]
@@ -1943,4 +1973,12 @@ class ClusterRefs:
                 mempool.free_all_blocks()
 
 
+class Accessed(NamedTuple):
+    column_index: int
+    cluster_start: int
+    cluster_stop: int
+    field_id: int
+    field_name: str
+
+
 uproot.classes["ROOT::RNTuple"] = Model_ROOT_3a3a_RNTuple
