@@ -462,7 +462,7 @@ class Tree:
         if self._num_baskets >= self._basket_capacity - 1:
             self._basket_capacity = max(
                 self._basket_capacity + 1,
-                int(math.ceil(self._basket_capacity * self._resize_factor)),
+                math.ceil(self._basket_capacity * self._resize_factor),
             )
 
             for datum in self._branch_data:
@@ -735,7 +735,7 @@ class Tree:
                             )
                         )
                     tofill.append((branch_name, datum["compression"], big_endian, None))
-                    if datum["kind"] == "counter":
+                    if datum["kind"] == "counter" and big_endian.size > 0:
                         datum["tleaf_maximum_value"] = max(
                             big_endian.max(), datum["tleaf_maximum_value"]
                         )
@@ -1098,22 +1098,20 @@ class Tree:
             out.append(uproot._util.tobytes(leaf_header))
             if len(leaf_name) < 255:
                 out.append(
-                    struct.pack(">B%ds" % len(leaf_name), len(leaf_name), leaf_name)
+                    struct.pack(f">B{len(leaf_name)}s", len(leaf_name), leaf_name)
                 )
             else:
                 out.append(
-                    struct.pack(
-                        ">BI%ds" % len(leaf_name), 255, len(leaf_name), leaf_name
-                    )
+                    struct.pack(f">BI{len(leaf_name)}s", 255, len(leaf_name), leaf_name)
                 )
             if len(leaf_title) < 255:
                 out.append(
-                    struct.pack(">B%ds" % len(leaf_title), len(leaf_title), leaf_title)
+                    struct.pack(f">B{len(leaf_title)}s", len(leaf_title), leaf_title)
                 )
             else:
                 out.append(
                     struct.pack(
-                        ">BI%ds" % len(leaf_title), 255, len(leaf_title), leaf_title
+                        f">BI{len(leaf_title)}s", 255, len(leaf_title), leaf_title
                     )
                 )
 
@@ -1397,6 +1395,13 @@ class Tree:
         fObjlen = len(uncompressed_data)
         fNbytes = fKeylen + len(compressed_data)
 
+        if max(fObjlen, fNbytes) > uproot.const.kMaxTBasketBytes:
+            raise ValueError(
+                f"Numpy array data of branch {branch_name} has an uncompressed size of {fObjlen} bytes "
+                f"and a compressed size of {fNbytes} bytes, which is too large to fit in a TBasket. "
+                "Neither of these sizes can exceed 2 GiB."
+            )
+
         parent_location = self._directory.key.location  # FIXME: is this correct?
 
         location = self._freesegments.allocate(fNbytes, dry_run=False)
@@ -1472,6 +1477,13 @@ class Tree:
         fObjlen = len(uncompressed_data)
         fNbytes = fKeylen + len(compressed_data)
 
+        if max(fObjlen, fNbytes) > uproot.const.kMaxTBasketBytes:
+            raise ValueError(
+                f"Jagged array data of branch {branch_name} has an uncompressed size of {fObjlen} bytes "
+                f"and a compressed size of {fNbytes} bytes, which is too large to fit in a TBasket. "
+                "Neither of these sizes can exceed 2 GiB."
+            )
+
         parent_location = self._directory.key.location  # FIXME: is this correct?
 
         location = self._freesegments.allocate(fNbytes, dry_run=False)
@@ -1561,6 +1573,13 @@ class Tree:
         fObjlen = len(uncompressed_data)
         fNbytes = fKeylen + len(compressed_data)
 
+        if max(fObjlen, fNbytes) > uproot.const.kMaxTBasketBytes:
+            raise ValueError(
+                f"String data of branch {branch_name} has an uncompressed size of {fObjlen} bytes "
+                f"and a compressed size of {fNbytes} bytes, which is too large to fit in a TBasket. "
+                "Neither of these sizes can exceed 2 GiB."
+            )
+
         parent_location = self._directory.key.location  # FIXME: is this correct?
 
         location = self._freesegments.allocate(fNbytes, dry_run=False)
